# Import the necessary modules and classes that define the package
from .SoundManager import SoundManager
from .ChatSpeechProcessor import ChatSpeechProcessor
from .ModuleLoader import ModuleLoader
from .Chat import Chat
from .ContextHandlers import ContextHandlers
from .ConnectionStatus import ConnectionStatus
from .LoadTts import LoadTts

# Define the package metadata
__name__ = 'daisy_llm'
__version__ = '0.0.4'
__author__ = 'Myra Krusemark'
__email__ = 'daisy_llm_tools@myrakrusemark.com'

# Export the classes and metadata for use by other modules and packages
__all__ = ['Chat', 'ContextHandlers', 'ModuleLoader', 'ChatSpeechProcessor', 'SoundManager', 'ConnectionStatus', 'LoadTts', '__name__', '__version__', '__author__', '__email__']
